/* parser.c
 * geg, a GTK+ Equation Grapher
 * David Bryant 1998
 */

/*
 * Corrected grammar
 * Kris Jurka 27/7/1999
 */

/*
 * Corrected more grammar
 * Torsten Landschoff 24/11/1999
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <math.h>
#include "parser.h"
#include "prefs.h"
#include "tokeniser.h"
#include "log.h"
#include "i18n.h"

/* #define DEBUG 1 */

/*
 *
 * EBNF:
 *  <expr>     ::= +|- <term> { +|- <term> }
 *  <term>     ::= <product> { *|/ <product> }
 *  <product>  ::= <factor> { ^ <factor> }
 *  <factor>   ::= <number> | <function> | (<expr>)
 *  <number>   ::= <float> | x
 *  <function> ::= <function_name>(<expression>)
 */

static parse_tree *split(void);
static parse_tree *expr(void);
static parse_tree *term(void);
static parse_tree *factor(void);
static parse_tree *product(void);

static token_list *list_g;      /* global list, makes life easier */
static gint parse_error;	/* flag this when a parse error occurs */

/* main variables a-z */
extern gdouble var[42];
extern parse_tree *fun[26];
extern gint varfun[26];
extern struct_ranges ranges;

/* some convenience functions follow, these stop seg faults and other stuff,
 * when the user enters un-parsable equations.
 * There is probably a fair bit of redundancy in the NULL pointer checking
 * but leave it there for now.
 */
static gint
token_type_is(gint type)
{
  if(!list_g)
    return(FALSE);
  return((list_g->type & 255) == type);
}

static gint
token_subtype()
{
  if(!list_g)
    return(-1);
  return(list_g->type >> 8);
}

static gint
token_type(void)
{
  if(!list_g)
    return(EOE);
  return(list_g->type);
}

static gdouble
token_value(void)
{
  if(!list_g)
    return(0);
  return(list_g->value);
}

static void
get_next_token(void)
{
  if(list_g)
    list_g = list_g->next;
}

static void
skip_token(gint type)
{
  if(!list_g) {
    parse_error = TRUE;
    return;
  }
  else
    if((list_g->type & 255)!= type)
      parse_error = TRUE;
  list_g = list_g->next;
}
/* end of convenience functions
 */

/* split all semicolon separated expressions
 */
static parse_tree *
split(void)
{
  parse_tree *tree = NULL, *temp_tree;
  while(!token_type_is(EOE)) {
    if (token_type_is(SEP)) {
      temp_tree = tree;
      tree = g_new(parse_tree, 1);
      tree->left = temp_tree;
      tree->token.type = token_type();
      get_next_token();
      if (!list_g) {
	tree->right = NULL;
	break;
      }
      tree->right = expr();
      if (!tree->right) break;
    } else {
      tree = expr();
      if (!tree) break;      
    }
  }
  return(tree);
}

/* expr, starting non-terminal for generating parse tree
 */
static parse_tree *
expr(void)
{
  parse_tree *tree = NULL, *temp_tree;

#if DEBUG  
  fprintf(stderr, "expr %d\n", token_type());
#endif

  if(token_type_is(DEF)) {
    get_next_token();
    if (token_type_is(EQU)) {
      tree = g_new(parse_tree, 1);
      tree->left = NULL;
      tree->token.type = token_type();
      get_next_token();
      tree->right = expr();
      return(tree);
    } else {
      parse_error = TRUE;
      return(NULL);
    }
  }

  if(token_type_is(NUM) ||
     token_type_is(VAR) ||
     token_type_is(FUN) ||
     token_type_is(SUB) ||	/* negate, (not minus) */
     token_type_is(LB)	||
     token_type_is(LCB) ||
     token_type_is(LSB)) {
    tree = term();
  }
  else {
    parse_error = TRUE;
    return(NULL);
  }

  while(token_type_is(ADD) ||
	token_type_is(SUB)) {
    temp_tree = tree;		/* save the old tree */
    tree = g_new(parse_tree, 1);
    tree->token.type = token_type();
    tree->left = temp_tree;	/* push old tree down a 'left branch' */
    get_next_token();
    tree->right = term();	/* 'right branch' */
  }

  if((tree == NULL) || (token_type_is(ERR)))
    parse_error = TRUE;
  return(tree);
}

/* term, non-terminal
 */
static parse_tree *
term(void)
{
  parse_tree *tree = NULL, *temp_tree;

#if DEBUG  
  fprintf(stderr, "term %d\n", token_type());
#endif

  if(token_type_is(NUM) ||
     token_type_is(VAR) ||
     token_type_is(FUN) ||
     token_type_is(SUB) ||	/* negate, (not minus) */
     token_type_is(LB)	||
     token_type_is(LCB) ||
     token_type_is(LSB)) {
    if(token_type_is(SUB)) {	/* negate */
      tree = g_new(parse_tree, 1);
      tree->token.type = SUB;	/* token_type(); */
      tree->left = g_new(parse_tree, 1);
      tree->left->token.type = NUM;
      tree->left->token.func = NULL;
      tree->left->token.value = 0;
      tree->left->left = NULL;
      tree->left->right = NULL;
      get_next_token();		/* skip the minus-sign */
      tree->right = product();
    }
    else {
      tree = product();
    }
  }
  else {
    parse_error = TRUE;
    return(NULL);
  }
  
  while(token_type_is(MUL) ||
	token_type_is(DIV) ||	
	token_type_is(VAR) ||
	token_type_is(FUN) ||
	token_type_is(LB)  ||
	token_type_is(LCB) ||
	token_type_is(LSB)) {
    if(token_type_is(VAR) ||	/* assume multiplication */
       token_type_is(FUN) ||
       token_type_is(LB)  ||
       token_type_is(LCB) ||
       token_type_is(LSB)) {
      temp_tree = tree;
      tree = g_new(parse_tree, 1);
      tree->token.type = MUL;   /* set multiplication */
      tree->left = temp_tree;
      tree->right = term();
    }
    else {
      temp_tree = tree;
      tree = g_new(parse_tree, 1);
      tree->token.type = token_type();
      tree->token.value = token_value();
      tree->left = temp_tree;
      get_next_token();
      tree->right = term();
    }
  }

  if(tree == NULL)
    parse_error = TRUE;
  return(tree);
}

/* product, non-terminal
 */
static parse_tree *
product(void)
{
  parse_tree *tree = NULL, *temp_tree;

#if DEBUG  
  fprintf(stderr, "prod %d\n", token_type());
#endif

  if(token_type_is(NUM) ||
     token_type_is(VAR) ||
     token_type_is(FUN) ||
     token_type_is(LB)	||
     token_type_is(LCB) ||
     token_type_is(LSB)) {
    tree = factor();
  }
  else {
    parse_error = TRUE;
    return(NULL);
  }

  while(token_type_is(POW)) {
    temp_tree = tree;
    tree = g_new(parse_tree, 1);
    tree->token.type = token_type();
    tree->left = temp_tree;
    get_next_token();
    tree->right = product();
  }

  if(tree == NULL)
    parse_error = TRUE;
  return(tree);
}

/* factor, non-terminal
 * negation is a bit of a hack, if we get a negator, we create a left tree,
 * with number=0, and a negation node, and continue down the right tree....
 * ie, we subtract the following factor from 0
 */
static parse_tree *
factor(void)
{
  parse_tree *tree;

#if DEBUG  
  fprintf(stderr, "factor %d\n", token_type());
#endif

  if(token_type_is(NUM) ||
     token_type_is(VAR)) {
    tree = g_new(parse_tree, 1);
    tree->left = NULL;
    tree->right = NULL;
    tree->token.type = token_type();
    tree->token.value = token_value();
    get_next_token();
    return(tree);
  }
  else if(token_type_is(FUN)) {
    tree = g_new(parse_tree, 1);
    tree->right = NULL;
    tree->token.type = token_type();
    tree->token.func = list_g->func;
    tree->token.value = 0;    
    skip_token(FUN);
    if(token_type_is(LB)) {
      skip_token(LB);
      tree->left = expr();
      skip_token(RB);
    }
    else if(token_type_is(LSB)) {
      skip_token(LSB);
      tree->left = expr();
      skip_token(RSB);
    }
    else {
      skip_token(LCB);
      tree->left = expr();
      skip_token(RCB);
    }
    return(tree);
  }
  else if(token_type_is(LB)) {
    skip_token(LB);
    tree = expr();
    skip_token(RB);
    return(tree);
  }
  else if(token_type_is(LSB)) {
    skip_token(LSB);
    tree = expr();
    skip_token(RSB);
    return(tree);
  }
  else if(token_type_is(LCB)) {
    skip_token(LCB);
    tree = expr();
    skip_token(RCB);
    return(tree);
  }
  else {
    parse_error = TRUE;
    return(NULL);
  }
}

static gint
fix_equal()
{
  token_list *list, *listp = NULL, *listpp;
  list = list_g;
  
  while(1) {
    /* search for x = ? tokens, and set variable to appropriate value 
       from previous token if it is a VAR preceded by a SEP, 
       otherwise do parse error */
    listpp = listp;
    listp = list;
    list = list->next;
    if (!list) break;
    if (list->type == EQU) {
      if (listpp && ((listpp->type & 255) != SEP)) {
        return(parse_error = TRUE);	
      }
      if ((listp->type & 255) != VAR) {
        return(parse_error = TRUE);
      }
      /* attach assigned variable to = operator */
      list->type |= listp->type & 0xff00;
      listp->type = DEF;
    }
  }
  return parse_error;
}

static parse_tree *
directive(void)
{
  token_list *list;
  parse_tree *tree = NULL;
  gint u, v, w;
  if (token_type_is(VAR)) {
    /* assignment of variable */
    list = list_g;
    get_next_token();
    if (!token_type_is(EQU)) return(NULL);
    return(make_parse_tree(list));
  }
  if (token_type_is(NUM) && list_g->func) {
    /* cart, polar, param, sequ */
    list = list_g;
    get_next_token();
    if (token_type_is(EOE)) {
      list->func(list->value);
      return((parse_tree *)list);
    }
    return(NULL);      
  }
  v = token_subtype() - 1;   
  if (token_type_is(FUN) && v >= 0 && v <= 25)
    /* assignment of user defined function ? */    
    get_next_token();
  else return(NULL);
  if (token_type_is(EQU)) {
    get_next_token();
    if(token_type_is(TGT)) {
      get_next_token();
      if(!token_type_is(RES)) return(NULL);
      get_next_token();
    }
    if(!token_type_is(FUN)) return(NULL);
    w = token_subtype() - 1;
    list = list_g;
    get_next_token();
    if(token_type_is(DER)) {
      get_next_token();
      if (!token_type_is(EOE)) return(NULL);
      if (fun[v]) { free_tree(fun[v]); fun[v] = NULL; }
      if (w >= 0) {
        /* derivative of user defined function */
        varfun[v] = -1-w;
        fun[v] = dup_tree(fun[w]);	
      } else {
        /* derivative of predefined function */
	if (!list->func) {
	  parse_error = TRUE;
	  return(NULL);
	}
        varfun[v] = -27;
	tree = g_new(parse_tree, 1);
	tree->left = NULL;
	tree->right = NULL;
	tree->token.type = FUN;
	tree->token.func = list->func;
        fun[v] = tree;
      }
      return(dup_tree(fun[v]));
    }
    if(token_type_is(RES)) {
      get_next_token();
      if (!token_type_is(VAR)) return(NULL);
      u = list_g->type >> 8;      
      get_next_token();
      if (!token_type_is(EOE)) return(NULL);      
      if (fun[v]) { free_tree(fun[v]); fun[v] = NULL; }
      if (w >= 0) {
        /* tangent of user defined function */
        varfun[v] = -28-w-(u << 8);
        fun[v] = dup_tree(fun[w]);
      } else {
        /* tangent of predefined function */
	if (!list->func) {
	  parse_error = TRUE;
	  return(NULL);
	}
        varfun[v] = -54-(u << 8);
	tree = g_new(parse_tree, 1);
	tree->left = NULL;
	tree->right = NULL;
	tree->token.type = FUN;
	tree->token.func = list->func;
        fun[v] = tree;
      }
      return(dup_tree(fun[v]));
    }    
    return(NULL);
  }
  
  if (token_type_is(LB)) get_next_token(); else return(NULL);
  if (token_type_is(VAR)) {
    w = token_subtype();
    get_next_token();      
  } else return(NULL);
  if (token_type_is(RB)) get_next_token(); else return(NULL);
  if (token_type_is(EQU)) get_next_token(); else return(NULL);
  if (token_type_is(EOE)) return(NULL);
  if (fix_equal()) return(NULL);
  tree = split();
  if((parse_error)||(!token_type_is(EOE))) return(NULL);
  if (fun[v]) free_tree(fun[v]);
  fun[v] = dup_tree(tree);
  varfun[v] = w;
  return(tree);  
}

/* make_parse_tree, creates a parse tree from a list of tokens
 */

parse_tree *
make_parse_tree(token_list *list)
{
  parse_tree *tree = NULL, *temp_tree;  
  parse_error = FALSE;
  list_g = list; 	/* assign global list */

  /* Check if initial token is #, ?, = or & */
  if (token_type_is(COM)) {
    /* comment */
    tree = g_new(parse_tree, 1);
    tree->left = NULL;
    tree->right = NULL;
    tree->token.type = COM;
    return(tree);
  } else
  if (token_type_is(ASK)) {
    /* evaluation request */
    get_next_token();
    if (fix_equal()) return(NULL);  
    temp_tree = split();
    if((!temp_tree) || (parse_error) || (!token_type_is(EOE))) return(NULL);
    tree = g_new(parse_tree, 1);
    tree->left = NULL;
    tree->right = temp_tree;
    tree->token.type = ASK;
    return(tree);    
  } else
  if (token_type_is(EQU)) {
    /* solving equation */
    get_next_token();
    if (fix_equal()) return(NULL);  
    temp_tree = split();
    if((!temp_tree) || (parse_error) || (!token_type_is(EOE))) return(NULL);
    tree = g_new(parse_tree, 1);
    tree->left = NULL;
    tree->right = temp_tree;
    tree->token.type = SOL;
    return(tree);    
  } else
  if (token_type_is(DEF)) {
    /* directive : change of coordinates or definition of function */
    get_next_token();  
    temp_tree = directive();
    if((!temp_tree) || (parse_error) || (!token_type_is(EOE))) return(NULL);    
    tree = g_new(parse_tree, 1);
    tree->left = NULL;
    tree->right = temp_tree;
    /* set value field from tree->right */
    tree->token.type = tree->right->token.type & 255;
    tree->token.value = (gdouble)
      (tree->token.type == EQU || tree->token.type == SEP);
    tree->token.type = DEF;    
    return(tree);    
  }

  if (fix_equal()) return(NULL);
  tree = split();
  if((parse_error)||(!token_type_is(EOE))) return(NULL);
  return(tree);
}

/* free_tree, frees the memory allocated to the tree using a recursive
 * algorithm
 */
void
free_tree(parse_tree *tree)
{
  if(tree->left)
    free_tree(tree->left);
  if(tree->right)
    free_tree(tree->right);
  g_free(tree);
  return;
}

/* dup_tree, duplicates the tree using a recursive algorithm
 */
parse_tree *
dup_tree(parse_tree *tree)
{
  parse_tree * new_tree;
  new_tree = g_new(parse_tree, 1);
  new_tree->token = tree->token;
  new_tree->right = (tree->right)? dup_tree(tree->right) : NULL;
  new_tree->left = (tree->left)? dup_tree(tree->left) : NULL;  
  return(new_tree);
}

gint
tan_check(gdouble c)
{
  gdouble x;
  x = var[(ranges.mode)? 't'-'a' : 'x'-'a'] - c;
  if (!isnan(var[TGTR]) && x > var[TGTR]) return 1;
  if (!isnan(var[TGTL]) && x < var[TGTL]) return 1;
  return 0;
}

/* eval_tree,       this function recursively evaluates the parse tree using
 * LRN notation. Currently only unary functions are supported, where the
 * argument to the function is assumed to be the left child branch of the
 * function node.... in the future binary functions might exist, like mod,
 * where the arguments are the children branches from left to right
 */
double
eval_tree(parse_tree *tree)
{
  gdouble x, v;
  gint i, j, k;
  Func func;

  /*
  printf("Tree[%d,%c] %d left %d right %d\n", 
         tree->token.type & 255, 'a'+ (tree->token.type >> 8),
         tree, tree->left, tree->right);
  */

  switch(tree->token.type & 255) {
  case ADD:
    return(eval_tree(tree->left) + eval_tree(tree->right));
  case SUB:
    return(eval_tree(tree->left) - eval_tree(tree->right));
  case MUL:
    return(eval_tree(tree->left) * eval_tree(tree->right));
  case DIV:
    x = eval_tree(tree->left) ; v = eval_tree(tree->right);
    if (v == 0) return(NAN);
    return(x / v);
  case POW:
    x = eval_tree(tree->left) ; v = eval_tree(tree->right);
    if (x < 0) return(NAN);
    if (x == 0) {
      if (v < 0) return(NAN);
      if (v == 0) return(1);
      if (v > 0) return(0);
    }
    return(pow(x,v));
  case SEP:
    (void) eval_tree(tree->left);
    return(eval_tree(tree->right));    
  case FUN:
    i = tree->token.type >> 8;
    if (i) {
      /* user defined function */
      --i;
      j = varfun[i];
      if (j >= 0) {
        x = var[j];
        var[j] = eval_tree(tree->left);
        if (fun[i]) v = eval_tree(fun[i]); else v = 0;
        var[j] = x;
        return(v);
      } else {
	k = (-j) & 255;
	if (k <= 26) {
	  j = k - 1;
	  k = varfun[j];
	  x = eval_tree(tree->left);
	  var[k] = x - var[INCR];
	  if (fun[j]) v = eval_tree(fun[j]); else v = 0;
	  var[k] = x + var[INCR];	  
	  if (fun[j]) v = var[MULT] * (eval_tree(fun[j]) - v); else v = 0;
	  return(v);
	} else
	if (k == 27) {
	  func = fun[i]->token.func;
	  x = eval_tree(tree->left);
	  v = var[MULT] * (func(x + var[INCR]) - func(x - var[INCR]));
	  return(v);
	} else
	if (k <= 53) {
	  j = -j - 28;
	  x = var[j >> 8];
	  j = j & 255;
	  if (var[ACTV] && tan_check(x)) return(NAN);
	  if (fun[j]) {
	    k = varfun[j];	    
	    var[k] = x - var[INCR];
	    v = eval_tree(fun[j]); 
	    var[k] = x + var[INCR];
	    v = var[MULT] * (eval_tree(fun[j]) - v);
	    var[k] = x;
	    v = v * (eval_tree(tree->left) - x) + eval_tree(fun[j]);
	  } else v = 0.0;
          return(v);
	} else
	if (k == 54) {
	  func = fun[i]->token.func;
	  x = var[(-j) >> 8];
	  if (var[ACTV] && tan_check(x)) return(NAN);	  
	  v = var[MULT] * (func(x + var[INCR]) - func(x - var[INCR]));
	  v = v * (eval_tree(tree->left) - x) + func(x);	  
	  return(v);	  
	}
	return(1);
      }
    } else {
      /* predefined function */
      return(tree->token.func(eval_tree(tree->left)));
    }      
  case NUM:
    return(tree->token.value);
  case VAR:
    return(var[tree->token.type>>8]);
  case EQU:
    x = eval_tree(tree->right);
    i = tree->token.type>>8;
    if (i < ENDVARIABLES) var[i] = x;
    return(x);
  default:
    write_log(_("Eval Error"));
    return(0);
  }
}

