/* tokeniser.c
 * geg, a GTK+ Equation Grapher
 * David Bryant 1998
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <math.h>

#include "tokeniser.h"
#include "localfunctions.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif /* M_PI */

#define NUMBER(c) (((c) >= '0') && ((c) <= '9'))

typedef struct {
  char *string;		/* the identifier string in the equation */
  gint type;		/* eg NUM, FUN */
  double value;		/* used if it is a number or a constant */
  Func func;		/* the address of the unary function to call */
} token_db;

static gint compare(token_db *a, gchar *b);

/* these must be in descending order of string length          */
static token_db tokens[] =
{ { "tgt_right",VAR|(28<<8),0,	NULL 	},  
  { "tgt_left", VAR|(27<<8),0,	NULL 	},
  { "der_step", VAR|(26<<8),0,	NULL 	},
  { "arcsin", FUN, 0,		asin	},
  { "arccos", FUN, 0,		acos	},
  { "arctan", FUN, 0,		atan	},
  { "arccsc", FUN, 0,		l_acsc	},
  { "arcsec", FUN, 0,		l_asec	},
  { "arccot", FUN, 0,		l_acot	},
  { "cotan",  FUN, 0,		l_cot,	},
  { "floor",  FUN, 0,		l_floor	},  
  { "gamma",  FUN, 0,		tgamma	},
  { "color",  VAR|(36<<8), 0,	NULL	},
  { "ptrad",  VAR|(37<<8), 0,	NULL	},
  { "polar",  NUM, 1,		l_coord	},
  { "param",  NUM, 2,		l_coord	},  
  { "sequ",   NUM, 3,		l_coord	},  
  { "cart",   NUM, 0,		l_coord	},
  { "amin",   VAR|(38<<8), 0,	NULL	},
  { "amax",   VAR|(39<<8), 0,	NULL	},
  { "nmin",   VAR|(40<<8), 0,	NULL	},
  { "tmin",   VAR|(40<<8), 0,	NULL	},
  { "nmax",   VAR|(41<<8), 0,	NULL	},  
  { "tmax",   VAR|(41<<8), 0,	NULL	},  
  { "asin",   FUN, 0,		asin	},
  { "acos",   FUN, 0,		acos	},
  { "atan",   FUN, 0,		atan	},
  { "acsc",   FUN, 0,		l_acsc	},
  { "asec",   FUN, 0,		l_asec	},
  { "acot",   FUN, 0,		l_acot	},
  { "sinc",   FUN, 0, 		l_sinc	},
  { "sinh",   FUN, 0,		sinh	},
  { "cosh",   FUN, 0,		cosh	},
  { "tanh",   FUN, 0,		tanh	},
  { "cbrt",   FUN, 0,		l_cbrt	},
  { "sqrt",   FUN, 0,		sqrt	},
  { "char",   FUN, 0,		l_char	},
  { "clip",   FUN, 0,		l_clip	},  
  { "sign",   FUN, 0,		l_sign	},
  { "cut",    FUN, 0,		l_cut	},
  { "tgt",    TGT, 0,		NULL	}, 
  { "nan",    NUM, NAN,		NULL	},
  { "out",    NUM, NAN,		NULL	},   /* result, index 33 */
  { "gap",    VAR|(34<<8), 0,	NULL	},  
  { "sin",    FUN, 0,		sin,	},
  { "cos",    FUN, 0,		cos,	},
  { "tan",    FUN, 0,		tan,	},
  { "csc",    FUN, 0,		l_csc	},
  { "sec",    FUN, 0,		l_sec	},
  { "cot",    FUN, 0,		l_cot	},
  { "exp",    FUN, 0,		exp	},  
  { "log",    FUN, 0,		log10	},
  { "abs",    FUN, 0,		fabs	},
  { "ln",     FUN, 0,		log	},
  { "lw",     VAR|(35<<8),0,	NULL 	},  
  { "pi",     NUM, M_PI,	NULL	},
  { "fa",     FUN, 0,		NULL	},
  { "fb",     FUN, 0,		NULL 	},
  { "fc",     FUN, 0,		NULL 	},
  { "fd",     FUN, 0,		NULL 	},
  { "fe",     FUN, 0,		NULL 	},
  { "ff",     FUN, 0,		NULL 	},  
  { "fg",     FUN, 0,		NULL 	},
  { "fh",     FUN, 0,		NULL 	},
  { "fi",     FUN, 0,		NULL 	},
  { "fj",     FUN, 0,		NULL 	},
  { "fk",     FUN, 0,		NULL 	},
  { "fl",     FUN, 0,		NULL 	},
  { "fm",     FUN, 0,		NULL 	},
  { "fn",     FUN, 0,		NULL 	},
  { "fo",     FUN, 0,		NULL 	},
  { "fp",     FUN, 0,		NULL 	},
  { "fq",     FUN, 0,		NULL 	},
  { "fr",     FUN, 0,		NULL 	},
  { "fs",     FUN, 0,		NULL 	},
  { "ft",     FUN, 0,		NULL 	},
  { "fu",     FUN, 0,		NULL 	},
  { "fv",     FUN, 0,		NULL 	},
  { "fw",     FUN, 0,		NULL 	},
  { "fx",     FUN, 0,		NULL 	},
  { "fy",     FUN, 0,		NULL 	},
  { "fz",     FUN, 0,		NULL 	},
  { ";",      SEP, 0,		NULL	},
  { "=",      EQU, 0,		NULL	},  
  { "+",      ADD, 0,		NULL	},
  { "-",      SUB, 0,		NULL	},
  { "*",      MUL, 0,		NULL	},
  { "/",      DIV, 0,		NULL	},
  { "^",      POW, 0,		NULL	},
  { "[",      LSB, 0,		NULL	},
  { "]",      RSB, 0,		NULL	},
  { "{",      LCB, 0,		NULL	},
  { "}",      RCB, 0,		NULL	},
  { "(",      LB,  0,		NULL	},
  { ")",      RB,  0,		NULL	},
  { "?",      ASK, 0,		NULL	},
  { "#",      COM, 0,		NULL	},  
  { "&",      DEF, 0,		NULL	},
  { "'",      DER, 0,		NULL	},
  { "_",      RES, 0,		NULL	},    
  { "a",      VAR, 0,		NULL	},
  { "b",      VAR, 0,		NULL	},
  { "c",      VAR, 0,		NULL	},
  { "d",      VAR, 0,		NULL	},
  { "e",      VAR, 0,		NULL	},
  { "f",      VAR, 0,		NULL	},
  { "g",      VAR, 0,		NULL	},
  { "h",      VAR, 0,		NULL	},
  { "i",      VAR, 0,		NULL	},
  { "j",      VAR, 0,		NULL	},
  { "k",      VAR, 0,		NULL	},
  { "l",      VAR, 0,		NULL	},
  { "m",      VAR, 0,		NULL	},
  { "n",      VAR, 0,		NULL	},
  { "o",      VAR, 0,		NULL	},
  { "p",      VAR, 0,		NULL	},
  { "q",      VAR, 0,		NULL	},
  { "r",      VAR, 0,		NULL	},
  { "s",      VAR, 0,		NULL	},
  { "t",      VAR, 0,		NULL	},
  { "u",      VAR, 0,		NULL	},
  { "v",      VAR, 0,		NULL	},
  { "w",      VAR, 0,		NULL	},
  { "x",      VAR, 0,		NULL	},
  { "y",      VAR, 0,		NULL	},
  { "z",      VAR, 0,		NULL	},  
  { "",       EOE, 0,		NULL	} };

/* output */

void
set_output(gdouble x)
{
  tokens[33].value = x;
}

/* compare, */
static gint
compare(token_db *a, gchar *b)
{
  if((*a->string == '\0') ^ (*b == '\0'))
    return(0);
  else
    return(!strncasecmp(a->string, b, strlen(a->string)));
}

/* make_token_list, this function builds a token list out of the equation   
 */
token_list *
make_token_list(gchar *equation)
{
  gint i, j;
  token_list *list;

  list = (token_list *)g_new(token_list, 1);

  /* skip any whitespace */
  while(*equation == ' ')
    equation++;

  /* check for the token in the token database */
  for(i = 0; i < (sizeof(tokens) / sizeof(token_db)); i++) {
    if(compare(&tokens[i], equation))
    {
      list->type = tokens[i].type;
      list->value = tokens[i].value;
      list->func = tokens[i].func;
      equation += (j = strlen(tokens[i].string));
      if (j == 1 && list->type == VAR)
        list->type |= ((tokens[i].string[0]-'a') << 8);
      else
      if (j == 2 && list->type == FUN && tokens[i].string[0] == 'f')
	/* byte 1 of list->type is given by a .. z -> 1 .. 26 */
        list->type |= (tokens[i].string[1] -'a' + 1) << 8;
      
      if(list->type != EOE)
        list->next = make_token_list(equation);
      else
        list->next = NULL;
      return(list);
    }
  }

  /* if it wasn't in the token database, it's either a number, or the token */
  /* can't be recognised and is therefore an error */
  if(NUMBER(*equation)) {
    list->type = NUM;
    list->value = 0;
    while(NUMBER(*equation)) {
      list->value = list->value * 10 +  *equation++ - '0';
    }
    if(*equation == '.') {	/* a real number */
      equation++;
      i = 0;		/* use i to count the decimal places */
      while(NUMBER(*equation))
        list->value += (*equation++ - '0') * pow(10, --i);
    }
  list->next = make_token_list(equation);
  return(list);
  }
  else {	/* it's an error .. */
    list->type = ERR;
    list->value = 0;
    list->next = NULL;
    return(list);
  }
}


/* free_list, a recursive list freeing algorithm  
 */
void free_list(token_list *list)
{
  if(list->next != NULL)
    free_list(list->next);
  g_free(list);
}
